// ST7781R TFT Library for mbed / NXP LPC1768
// (c) 2011 proteino.de (Norbert Schroeder)
// Licence: LGPL 2.1 or above
//
// Original Arduino-version:
//  2011 Copyright (c) Seeed Technology Inc.
//  Authors: Albert.Miao, Visweswara R (with initializtion code from TFT vendor)
//  Licence: LGPL 2.1 or above
///////////////////////////////////////////////////////////////////////////

// Bus-Definitions:
//   DataBus is on Port0.4-11 (mbed-Pins: 30/29/8/7/6/5/28/27)
//   Control Pins on Port2.2-5
//		- ctrl_CS Port2.5, mbed p21 - 0=enable, 1=disable
//		- ctrl_RS Port2.4, mbed p22 - alias DCX, 0=cmd, 1=data
//		- ctrl_WR Port2.3, mbed p23
//		- ctrl_RD Port2.2, mbed p24
//   Touch Pins
//		- X+ = GND (or mbed p18 with 4-wire-functions)
//		- Y+ = mbed p17
//		- X- = mbed p16
//		- Y- = VCC (or mbed p15 with 4-wire-functions)

#ifndef TFT_h
#define TFT_h

#define LPC17xx

//Some Colors
#define RED			0xf800
#define GREEN		0x7e00
#define BLUE		0x001f
#define BLACK		0x0000
#define YELLOW		0xffe0
#define WHITE		0xffff
#define CYAN		0x07ff
#define BRIGHT_RED	0xf810
#define GRAY1		0x8410
#define GRAY2		0x4208
#define BTNLIGHT	0xC618
#define BTNFACE		0x8410
#define BTNSHADOW	0x4208
#define FGCOLOR		0x0000

//TFT resolution 240*320
#define MIN_X	0
#define MIN_Y	0
#define MAX_X	240
#define MAX_Y	320

#ifdef LPC17xx

#define PORT0_MASK		0x00000FF0U
#define PORT0_MASK_DB	0x00000FF0U
#define PORT2_MASK		0x0000003CU
#define	ctrl_CS			0x00000020U
#define ctrl_RS			0x00000010U
#define ctrl_WR			0x00000008U
#define ctrl_RD			0x00000004U

//========================================
#define CS_BIT			0x00000001U
#define CS_HIGH			*GPIO2_SET=ctrl_CS;
#define CS_LOW			*GPIO2_CLR=ctrl_CS;
//-----------------------------------------
#define RS_BIT			0x00000002U
#define RS_HIGH			*GPIO2_SET=ctrl_RS;
#define RS_LOW			*GPIO2_CLR=ctrl_RS;

//-----------------------------------------
#define WR_BIT			0x00000004U
#define WR_HIGH			*GPIO2_SET=ctrl_WR;
#define WR_LOW			*GPIO2_CLR=ctrl_WR;
//#define WR_RISING		{PORT_WR|=WR_BIT;PORT_WR&=~WR_BIT;}
//-----------------------------------------
#define RD_BIT			0x00000008U
#define RD_HIGH			*GPIO2_SET=ctrl_RD;
#define RD_LOW			*GPIO2_CLR=ctrl_RD;
//#define RD_RISING		{PORT_RD|=RD_BIT;PORT_RD&=~RD_BIT;}
//========================================

#endif

#include "../system/LPC17xx_system.h"
#include "system/_LPC17xx_register.h"
#include "../lcd/font.c"

extern const unsigned char simpleFont[][8];
void 			pushData			(unsigned char a);
void			all_pin_input		(void);
void			all_pin_output		(void);
void			all_pin_low			(void);
void 			sendData			(unsigned int data);
void			sendCommand			(unsigned int index);
unsigned char 	getData				(void);
unsigned int	readRegister		(unsigned int index);
void			LCD_Init			(void);
void			LCD_ExitStandBy		(void);
void			LCD_SetOrientation	(unsigned int HV);
void			LCD_SetXY			(unsigned int poX, unsigned int poY);
void			LCD_ClrScr			(unsigned int c);
void			LCD_SetPixel		(unsigned int poX, unsigned int poY,unsigned int color);
void			LCD_Circle			(int poX, int poY, int r,unsigned int color);
void			LCD_FillCircle		(int poX, int poY, int r,unsigned int color);
void			LCD_Line			(int x0,int y0,int x1,int y1,unsigned int color);
void			LCD_VerticalLine	(unsigned int poX, unsigned int poY,unsigned int length,unsigned int color);
void			LCD_HorizontalLine	(unsigned int poX, unsigned int poY,unsigned int length,unsigned int color);
void			LCD_Rectangle		(unsigned int poX, unsigned int poY, unsigned int length,unsigned int width,unsigned int color);
void			LCD_FillRectangle	(unsigned int poX, unsigned int poY, unsigned int length,unsigned int width,unsigned int color);
void			LCD_Char			(unsigned char ascii,unsigned int poX, unsigned int poY,unsigned int size, unsigned int fgcolor);
void			LCD_String			(char *string,unsigned int poX, unsigned int poY,unsigned int size,unsigned int fgcolor);
uint32_t		LCD_Button			(unsigned int x,unsigned int y,char *caption);
void			LCD_TestScreen		(unsigned int a);
uint32_t		Touch_ReadX			(void);
uint32_t		Touch_ReadY			(void);
uint32_t		Touch2wire_ReadX	(void);
uint32_t		Touch2wire_ReadY	(void);


void pushData(unsigned char a){
 all_pin_low();
#ifdef LPC17xx
 *GPIO0_PIN=(a&0xFF)<<4;
#endif
}

void all_pin_input(void){
#ifdef LPC17xx
 *GPIO0_MASK=~PORT0_MASK;					// Port0-Maske, nur DatenBus
 *GPIO0_DIR=~PORT0_MASK;					// Port0 DB als Ausgang
#endif
}

void all_pin_output(void){
#ifdef LPC17xx
 *GPIO0_MASK=~PORT0_MASK;					// Port0-Maske, nur DatenBus
 *GPIO0_DIR=PORT0_MASK;						// Port0 DB als Ausgang
#endif
}

void all_pin_low(void){
#ifdef LPC17xx
 *GPIO0_CLR=(0xFF)<<4;
#endif
}

void sendData(unsigned int data){
    CS_LOW;
    RS_HIGH;
    RD_HIGH;
    WR_LOW;
    pushData((data&0xff00)>>8);
    WR_HIGH;
    WR_LOW;
    pushData(data&0xff);
    WR_HIGH;
    CS_HIGH;
/*
 *GPIO2_PIN=ctrl_RS|ctrl_RD;						// CLR=CS|WR;SET RS|RD
 *GPIO2_CLR=ctrl_CS|ctrl_WR;
// *GPIO0_CLR=(0xFF)<<4;*GPIO0_SET=(data&0xFF00)>>4;
 pushData((data&0xff00)>>8);
 *GPIO2_SET=ctrl_WR;
 *GPIO2_CLR=ctrl_WR;
    pushData(data&0xff);
// *GPIO0_CLR=(0xFF)<<4;*GPIO0_SET=(data&0xFF)<<4;__NOP();
 *GPIO2_SET=ctrl_WR;
 *GPIO2_SET=ctrl_CS;
*/
}

void sendCommand(unsigned int index){
    CS_LOW;
    RS_LOW;
    RD_HIGH;
    WR_HIGH;
    WR_LOW;
    pushData(0);
    WR_HIGH;
    WR_LOW;
    pushData(index&0xff);
    WR_HIGH;
    CS_HIGH;

/*	// following code seems to be same but doesn't work, even additional NOPs are not a help - mpf
 *GPIO2_PIN=ctrl_RD;			// RS=0 -> command data ; CLR CS|RS|WR;SET RD
// *GPIO0_CLR=(0xFF)<<4;
 *GPIO0_PIN=0;
 *GPIO2_SET=ctrl_WR;
 *GPIO2_CLR=ctrl_WR;
 *GPIO0_CLR=(0xFF)<<4;
 *GPIO0_PIN=(index&0xFF)<<4;
//    pushData(index&0xff);
 *GPIO2_SET=ctrl_WR;
 *GPIO2_SET=ctrl_CS;
*/
}

unsigned char getData(void){
 unsigned char data;
 sysPause(1);
 data = *GPIO0_PIN>>4;
 return data;
}

unsigned int readRegister(unsigned int index){
    unsigned int data=0;

    CS_LOW;
    RS_LOW;
    RD_HIGH;

    all_pin_output();

    WR_LOW;
    pushData(0);
    WR_HIGH;

    WR_LOW;
    pushData(index);
    WR_HIGH;

    all_pin_input();
    all_pin_low();
    RS_HIGH;

    RD_LOW;
    RD_HIGH;
    data |= getData()<<8;

    RD_LOW;
    RD_HIGH;
    data |= getData();

    CS_HIGH;
    all_pin_output();
    return data;
}

void LCD_Init(void){
 *GPIO0_MASK=~PORT0_MASK;					// Port0-Mask, DataBus only
 *GPIO0_DIR=PORT0_MASK;						// Port0 DB as Output
 *GPIO2_MASK=~PORT2_MASK;					// Port2-Mask, ControlBus only
 *GPIO2_DIR=PORT2_MASK;						// Port2 DB as Output
 *GPIO0_CLR=PORT0_MASK;
// *GPIO2_CLR=ctrl_CS;

    sysPause(100);

    sendCommand(0x0001);
    sendData(0x0100);
    sendCommand(0x0002);
    sendData(0x0700);
    sendCommand(0x0003);
    sendData(0x1030);
    sendCommand(0x0004);
    sendData(0x0000);
    sendCommand(0x0008);
    sendData(0x0302);
    sendCommand(0x000A);
    sendData(0x0000);
    sendCommand(0x000C);
    sendData(0x0000);
    sendCommand(0x000D);
    sendData(0x0000);
    sendCommand(0x000F);
    sendData(0x0000);

    sysPause(100);

    sendCommand(0x0030);
    sendData(0x0000);
    sendCommand(0x0031);
    sendData(0x0405);
    sendCommand(0x0032);
    sendData(0x0203);
    sendCommand(0x0035);
    sendData(0x0004);
    sendCommand(0x0036);
    sendData(0x0B07);
    sendCommand(0x0037);
    sendData(0x0000);
    sendCommand(0x0038);
    sendData(0x0405);
    sendCommand(0x0039);
    sendData(0x0203);
    sendCommand(0x003c);
    sendData(0x0004);
    sendCommand(0x003d);
    sendData(0x0B07);
    sendCommand(0x0020);
    sendData(0x0000);
    sendCommand(0x0021);
    sendData(0x0000);
    sendCommand(0x0050);
    sendData(0x0000);
    sendCommand(0x0051);
    sendData(0x00ef);
    sendCommand(0x0052);
    sendData(0x0000);
    sendCommand(0x0053);
    sendData(0x013f);

    sysPause(100);

    sendCommand(0x0060);
    sendData(0xa700);
    sendCommand(0x0061);
    sendData(0x0001);
    sendCommand(0x0090);
    sendData(0x003A);
    sendCommand(0x0095);
    sendData(0x021E);
    sendCommand(0x0080);
    sendData(0x0000);
    sendCommand(0x0081);
    sendData(0x0000);
    sendCommand(0x0082);
    sendData(0x0000);
    sendCommand(0x0083);
    sendData(0x0000);
    sendCommand(0x0084);
    sendData(0x0000);
    sendCommand(0x0085);
    sendData(0x0000);
    sendCommand(0x00FF);
    sendData(0x0001);
    sendCommand(0x00B0);
    sendData(0x140D);
    sendCommand(0x00FF);
    sendData(0x0000);
    sysPause(100);
    sendCommand(0x0007);
    sendData(0x0133);
    sysPause(50);
    LCD_ExitStandBy();
	LCD_ClrScr(0);
}

void LCD_ExitStandBy(void){
 sendCommand(0x0010);
 sendData(0x14E0);
 sysPause(100);
 sendCommand(0x0007);
 sendData(0x0133);
}

void LCD_SetOrientation(unsigned int HV){		// horizontal or vertical
 sendCommand(0x03);
 if(HV==1)sendData(0x5038);						// vertical
 else sendData(0x5030);							// horizontal
 sendCommand(0x0022);							// Start to write to display RAM
}

void LCD_SetXY(unsigned int poX, unsigned int poY){
 sendCommand(0x0020);//X
 sendData(poX);
 sendCommand(0x0021);//Y
 sendData(poY);
 sendCommand(0x0022);//Start to write to display RAM
}

void LCD_ClrScr(unsigned int c){				// c is backgroundcolor
 LCD_SetXY(0,0);
// for(uint32_t a=0;a<320*240;a++)sendData(c);	// the shorter code needs 27s for 256 CLS
 CS_LOW;
 RS_HIGH;
 RD_HIGH;
 uint32_t ch,cl;
 ch=(c&0xff00)>>4;
 cl=(c&0xff)<<4;

//  for(uint32_t a=0;a<320*240;a++){			// this shorter code needs 17s for 256 CLS
//   WR_LOW;pushData(ch);WR_HIGH;
//   WR_LOW;pushData(cl);WR_HIGH;
//  }

 if(ch!=cl){									// But how fast can you dance? :)
  for(uint32_t a=0;a<320*240;a++){				// 5.53s for 256 CLS - that's ok
   __NOP();__NOP();
   __NOP();__NOP();
   WR_LOW;
   __NOP();__NOP();
   *GPIO0_CLR=0xFF0;
   *GPIO0_PIN=ch;
   __NOP();__NOP();
   WR_HIGH;
   __NOP();__NOP();
   __NOP();__NOP();
   WR_LOW;
   __NOP();__NOP();
   *GPIO0_CLR=0xFF0;
   *GPIO0_PIN=cl;
   __NOP();__NOP();
   WR_HIGH;
  }
 }else{											// ClrScr with cl=ch, mostly used by black or white background
  for(uint32_t a=0;a<320*240;a++){				// 2.45s for 256 CLS - that's really fast!
   WR_LOW;
   *GPIO0_CLR=0xFF0;
   *GPIO0_PIN=cl;
   WR_HIGH;
   __NOP();
   WR_LOW;
   *GPIO0_CLR=0xFF0;
   *GPIO0_PIN=cl;
   WR_HIGH;
  }
 }
 CS_HIGH;
}

void LCD_SetPixel(unsigned int poX, unsigned int poY,unsigned int c){
 LCD_SetXY(poX,poY);
 sendData(c);
/*
 CS_LOW;
 RS_HIGH;
 RD_HIGH;
 uint32_t ch,cl;
 ch=(c&0xff00)>>4;
 cl=(c&0xff)<<4;

   WR_LOW;
   __NOP();__NOP();
   *GPIO0_CLR=0xFF0;
   *GPIO0_PIN=ch;
   __NOP();__NOP();
   WR_HIGH;
   __NOP();__NOP();
   __NOP();__NOP();
   WR_LOW;
   __NOP();__NOP();
   *GPIO0_CLR=0xFF0;
   *GPIO0_PIN=cl;
   __NOP();__NOP();
   WR_HIGH;
 CS_HIGH;
*/
}

void LCD_Circle(int poX, int poY, int r,unsigned int color){
    int x = -r, y = 0, err = 2-2*r, e2; 
    do {
        LCD_SetPixel(poX-x, poY+y,color); 
        LCD_SetPixel(poX+x, poY+y,color); 
        LCD_SetPixel(poX+x, poY-y,color); 
        LCD_SetPixel(poX-x, poY-y,color); 
        e2 = err;
        if (e2 <= y) { 
            err += ++y*2+1;
            if (-x == y && e2 <= x) e2 = 0; 
        }
        if (e2 > x) err += ++x*2+1; 
    } while (x <= 0);
}

void LCD_FillCircle(int poX, int poY, int r,unsigned int color){
 int x = -r, y = 0, err = 2-2*r, e2;
 do{
  LCD_VerticalLine(poX-x,poY-y,2*y,color);
  LCD_VerticalLine(poX+x,poY-y,2*y,color);
  e2 = err;
  if (e2 <= y){
   err += ++y*2+1;
   if (-x == y && e2 <= x) e2 = 0;
  }
  if (e2 > x) err += ++x*2+1;
 } while (x <= 0);
}


void LCD_Line(int x0,int y0,int x1,int y1,unsigned int color){

    int dx = abs(x1-x0), sx = x0<x1 ? 1 : -1;
    int dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1;
    int err = dx+dy, e2; // error value e_xy
/*
debugf("x0,y0 x1,y1 %i,%i %i,%i\n",x0,y0,x1,y1);
debugf("dx=%i\n",dx);
debugf("dy=%i\n",dy);
debugf("sx=%i\n",sx);
debugf("sy=%i\n\n",sy);
*/
    for (;;){ // loop
        LCD_SetPixel(x0,y0,color);
        e2 = 2*err;
        if (e2 >= dy) { // e_xy+e_x > 0
            if (x0 == x1) break;
            err += dy; x0 += sx;
        }
        if (e2 <= dx) { // e_xy+e_y < 0
            if (y0 == y1) break;
            err += dx; y0 += sy;
        }
    }

/*
 uint32_t x,y,dx,dy;
 if(x0>x1){
  x=x0;x0=x1;x1=x;
  x=y0;y0=y1;y1=x;
 }
 dx=abs(x1-x0);
 dy=abs(y1-y0);
 if(dx>dy){				// more dots in x
  for(x=x0;x<=x1;x++){
  
  }
 }else{					// more dots in y
  for(y=y0;y<=y1;y++){
  }
 }
*/
}


void LCD_VerticalLine(unsigned int poX, unsigned int poY,unsigned int length,unsigned int c){
 LCD_SetXY(poX,poY);
 LCD_SetOrientation(1);
 if(length+poY>MAX_Y)length=MAX_Y-poY;
/*
 CS_LOW;
 RS_HIGH;
 RD_HIGH;
 for(unsigned int i=0;i<length;i++){
  WR_LOW;pushData((c&0xff00)>>8);WR_HIGH;
  WR_LOW;pushData(c&0xff);WR_HIGH;
 }
 CS_HIGH;
*/
 for(unsigned int i=0;i<length;i++)sendData(c);
}

void LCD_HorizontalLine(unsigned int poX, unsigned int poY,unsigned int length,unsigned int color){
 LCD_SetXY(poX,poY);
 LCD_SetOrientation(0);
 if(length+poX>MAX_X)length=MAX_X-poX;
 for(unsigned int i=0;i<length;i++)sendData(color);
}

void LCD_Rectangle(unsigned int poX, unsigned int poY, unsigned int length,unsigned int width,unsigned int color){
 LCD_HorizontalLine(poX, poY, length, color);
 LCD_HorizontalLine(poX, poY+width, length, color);
 LCD_VerticalLine(poX, poY, width,color);
 LCD_VerticalLine(poX + length, poY, width,color);
}

void LCD_FillRectangle(unsigned int poX, unsigned int poY, unsigned int length, unsigned int width, unsigned int color){
 for(unsigned int i=0;i<width;i++)LCD_HorizontalLine(poX, poY+i, length, color);
}

void LCD_Char(unsigned char ascii,unsigned int poX, unsigned int poY,unsigned int size, unsigned int fgcolor){
 LCD_SetXY(poX,poY);
 LCD_SetOrientation(1);
 if((ascii < 0x20)||(ascii > 0x7e))ascii='?';			//Unsupported char.
 for(unsigned char i=0;i<8;i++){
  unsigned char temp = simpleFont[ascii-0x20][i];
  for(unsigned char f=0;f<8;f++){
   if((temp>>f)&0x01)LCD_FillRectangle(poX+i*size, poY+f*size, size, size, fgcolor);
  }
 }
}

void LCD_String(char *string,unsigned int poX, unsigned int poY,unsigned int size,unsigned int fgcolor){
 while(*string){
  for(unsigned char i=0;i<8;i++)LCD_Char(*string, poX, poY, size, fgcolor);
  *string++;
  if(poX < MAX_X){
   poX+=8*size; // Move cursor right
  }
 }
}
/*
void LCD_Integer(uint32_t number,unsigned int poX, unsigned int poY,unsigned int size,unsigned int fgcolor){
 uint32_t i;
 buf[5]=0;
 for(i=4;i!=0-1;i--){
  buf[i]=number%10+'0';
  if(!(number/=10))break;
 }
 while(*buf){
  for(unsigned char i=0;i<8;i++)LCD_Char(*buf, poX, poY, size, fgcolor);
  *buf++;
  if(poX < MAX_X){
   poX+=8*size; // Move cursor right
  }
 }
}
*/
uint32_t LCD_Button(unsigned int x,unsigned int y,char *caption){	// returns the width of the button
 uint32_t w=0,h=15;
 while(caption[w])w++;
 w=w*8+4;
 LCD_HorizontalLine(x  ,y  ,w      ,BTNLIGHT);
 LCD_HorizontalLine(x  ,y+h,w      ,BTNSHADOW);
 LCD_VerticalLine  (x  ,y  ,h      ,BTNLIGHT);
 LCD_VerticalLine  (x+w,y  ,h      ,BTNSHADOW);
 LCD_FillRectangle (x+1,y+1,w-2,h-2,BTNFACE);
 LCD_String(caption,x+3,y+3,1,FGCOLOR);
 return w;
}

void LCD_TestScreen(unsigned int a){
 LCD_ClrScr(0);
 LCD_FillRectangle(0,a,30,10,BLACK); 
 LCD_FillRectangle(30,a,30,10,RED);
 LCD_FillRectangle(60,a,30,10,GREEN);
 LCD_FillRectangle(90,a,30,10,BLUE);
 LCD_FillRectangle(120,a,30,10,CYAN);
 LCD_FillRectangle(150,a,30,10,YELLOW);
 LCD_FillRectangle(180,a,30,10,WHITE);
 LCD_FillRectangle(210,a,30,10,GRAY1);
 LCD_Rectangle(180,31,30,42,0x012C);
 LCD_Rectangle(181,32,28,40,0x0FF0);
 LCD_Circle(90,130,10,0xFF00);
 LCD_Circle(90,190,10,0x00FF);
 LCD_Circle(150,190,10,0x0FF0);
 LCD_Circle(150,130,10,0xF00F);
 LCD_FillCircle(119,159,15,BLUE);
 LCD_VerticalLine(119,130,60,0xFFF0);
 for(uint32_t x=0;x<100;x+=10){
  LCD_HorizontalLine(100,x,60,0xFFFF);
  LCD_VerticalLine(x,100,60,0xFFFF);
 }
 LCD_HorizontalLine(89,159,60,0x0FFF);
 LCD_HorizontalLine(89,119,60,0x0FFF);
 LCD_Line(99,139,139,179,0xFFFF);
 LCD_Line(139,139,99,179,0xFFFF);
 LCD_Line(0,0,239,319,0xFF00);
 LCD_Line(239,0,0,319,0xFF00);
 LCD_Line(239,0,0,159,0xF008);
 LCD_Char('S',10,10,2,RED);
 LCD_String("Test",10,30,2,0xF00F-a);
 a=LCD_Button(0,300,"TestBtn1");
 a+=LCD_Button(a+1,300,"TestBtn2");
 a+=LCD_Button(a+1,300,"gjBtn3");
}

uint32_t Touch_ReadX(void){
 // xp = Output Low (Sink)
 // xm = Output High (Source)
 // yp = Tri-state (Hi-Z), Input no Pull-up
 // yp = Tri-state (Hi-Z), Input no Pull-up -> wait -> ADC
 // (1<<20)|(1<<18)|(1<<16)|(1<<14);				// AD0.0-3
// uint32_t tpout=1
// *GPIO1_DIR=p1;*GPIO1_MASK=~p1;
}

uint32_t Touch_ReadY(void){
}

uint32_t Touch2wire_ReadX(void){
 uint32_t a;
 ADC_init(1<<1);
 a=ADC_read();
 ADC_free();
// LCD_String("Touch X-: ",10,200,1,0xF88F);
// LCD_String(int2str(a),100,200,1,0xF88F);
 if((a>0)&&(a<1160))a=(1160-a)*240/1160;else a=0;	// probably a Touch
 return a;
}

uint32_t Touch2wire_ReadY(void){
 uint32_t a;
 ADC_init(1<<2);
 a=ADC_read();
 ADC_free();
// LCD_String("Touch Y+: ",10,210,1,0xF88F);
// LCD_String(int2str(a),100,210,1,0xF88F);
 if(a<4000){													// ADC-Values in the area of 2250-4000 (resolution 1750), probably a Touch
  a=4000-a;
  if(a>1750)a=1750;
  a=a*320/1750;
 }else a=0;
 return a;
}

#endif

