// unit for analog digital conversion with LPC17xx
// designed for ARM Cortex-M3 from NXP (LPC*)
// (c) 2011 proteino.de (Norbert Schroeder)
// Licence: LGPL 2.1 or above

////////////// prototype begin ////////////////////////////////////////////////
void ADC_init(uint32_t adc);							// inits AD0.adc-Pins - only 1 Pin at a time for following software-ADC-read
void ADC_free(void);									// frees the AD0.x-Pins
void ADC_deactive(void);								// Save energy and deactivate the power for ADC
uint32_t ADC_read(void);								// read AD0.5 - LPC1768 p1.31 (mbed p20)
uint32_t LDR5_read(void);								// reads LDR on AD0.5, stores it in an array
uint32_t LDR_interpret(void);							// is called if switch pressed within LDR5_read, output of interpreted light-value
////////////// prototype end //////////////////////////////////////////////////

////////////// ADC begin //////////////////////////////////////////////////////
// Quellen: UserManual Chapter 29 (S. 574)
// http://www.keil.com/forum/17115/
// http://www.embeddedrelated.com/groups/lpc2000/show/43628.php
static int aLDR[10];						// aLDR[0] ist aktuelle Position im Helligkeits-Array, Werte [1..8] liegen zw. 0-4095, [9] ist Durchschnitt

void ADC_init(uint32_t adc){				// Aktivierung AD0.x with x=adc-Pins 0-7 - LPC1768 p1.31 (mbed p20)
 uint32_t adP0=0,adP1=0,adP3=0;				// Hilfsvariablen fr Pinsel0/1/3
 *PCONP |= (1<<12);							// Enable power to ADC block (PCADC-Bit)
 *AD0CR=adc|(4<<8)|(1<<21);//|(1<<24);		// S. 577: AD0.x; 4=0b100, also CLKDIV=5; PDN normal; Start conversion now auskommentiert
 if(adc&(1<<6))adP0|=(0x2<<6);				// PINSEL0 p0.3 in Bits 7:6 auf 0b10 fr AD0.6 - wenn Fkt<>0b11 das Doppelbit zuvor per & lschen! Woher hatte ich die Info? ^^
 if(adc&(1<<7))adP0|=(0x2<<4);				// PINSEL0 p0.2 in Bits 5:4 auf 0b10 fr AD0.7
 *PINSEL0 &= ~adP0;
 *PINSEL0 |= adP0;
 if(adc&(1<<0))adP1 =(0x1<<14);				// PINSEL0 p0.23 in Bits 15:14 auf 0b01 fr AD0.0 (mbed p15)
 if(adc&(1<<1))adP1|=(0x1<<16);				// PINSEL0 p0.24 in Bits 17:16 auf 0b01 fr AD0.1 (mbed p16)
 if(adc&(1<<2))adP1|=(0x1<<18);				// PINSEL0 p0.25 in Bits 19:18 auf 0b01 fr AD0.2 (mbed p17)
 if(adc&(1<<3))adP1|=(0x1<<20);				// PINSEL0 p0.26 in Bits 21:20 auf 0b01 fr AD0.3 (mbed p18)
 *PINSEL1 &= ~adP1;
 *PINSEL1 |= adP1;
 if(adc&(1<<4))adP3=(0x3<<28);				// PINSEL3 p1.30 in Bits 29:28 auf 0b11 fr AD0.4 (mbed p19)
 if(adc&(1<<5))adP3|=(0x3<<30);				// PINSEL3 p1.31 in Bits 31:30 auf 0b11 fr AD0.5 (mbed p20)
 *PINSEL3 &= ~adP3;
 *PINSEL3 |= adP3;
}

void ADC_free(void){						// Gibt alle AD0.x-Pins wieder frei
 *AD0CR&=~(0xFF);							// S. 577: alle AD0.x-Pins aus der Liste der AD-Pins werfen
 *PINSEL3 &= ~(0x3<<30);					// PINSEL3 p1.31 in Bits 31:30 auf 0b00 fr GPIO
}

void ADC_deactive(void){					// Save energy and deactivate the power for ADC
 *AD0CR&=~(1<<21);							// S. 574: PDN (Power Down)
 *PCONP &= ~(1<<12);						// Disable power to ADC block (PCADC-Bit)
}

uint32_t ADC_read(void){
 uint32_t regVal;
 *AD0CR |= (1<<24);							// Start conversion now
 while(1){									// wait until end of A/D convert
  regVal = *AD0GDR;
  if((regVal>>31)&1)break;
 }
 *AD0CR &= 0xF8FFFFFF;						// stop ADC now
 if((regVal>>30)&1)return(10000);			// Overrun
 regVal=(regVal>>4)&0xFFF;					// RESULT is on bits 15:4
 return(regVal); 
}

uint32_t LDR5_read(void){					// LDR on AD0.5
  uint32_t regVal;
  regVal=ADC_read();
  if(regVal<4096){							// Overrun returns 10000, 3.3V gives about 4080
   if(debLevel==3)debugf("%i\n",regVal);
   aLDR[aLDR[0]]=regVal;
   if(aLDR[0]<8)aLDR[0]++;else aLDR[0]=1;	// aLDR-Zeiger hochschieben bzw. beim Ende des Arrays zurck auf 1
  }
 if(SW_pressed()==1)return LDR_interpret();
 return 10000;
}

uint32_t LDR_interpret(void){				// gives a light-value for the LDR
 int x;
 aLDR[9]=0;									// Durchschnittswert der Helligkeit auf 0 setzen
 for(x=1;x<9;x++)aLDR[9]+=aLDR[x];			// ergibt also max 8*4096=2^15
 aLDR[9]>>=11;								// Auswertung auf Werte [0..15] = 2^4 skaliert, also SHR 11
 if(debLevel>=2){
  debugf("Helligkeit: ");
  if(aLDR[9]==0)debugf("Da sitzt was auf dem Sensor.\n");
  else if(aLDR[9]==1)debugf("dunkel\n");
  else if(aLDR[9]<=3)debugf("halbdunkel\n");
  else if(aLDR[9]<=8)debugf("durschnittlich\n");
  else if(aLDR[9]<=13)debugf("hell\n");
  else if(aLDR[9]==14)debugf("sehr hell\n");
  else if(aLDR[9]==15)debugf("Der Sensor fuehlt sich geblendet.\n");
 }
 return aLDR[9];
}
////////////// ADC end ////////////////////////////////////////////////////////
