// System-Calls, optimized dor Cortex-M-Series from NXP (tested only with LCP1768/mbed)
// (c) 2011 proteino.de (Norbert Schroeder)
// Licence: LGPL 2.1 or above

#ifndef LPC17xx_system_h
#define LPC17xx_system_h

#include "system/_LPC17xx_register.h"

static inline void __enable_irq()			{ __asm volatile ("cpsie i"); }
static inline void __disable_irq()			{ __asm volatile ("cpsid i"); }
static inline void __NOP()					{ __asm volatile ("nop"); }
static inline void __WFI()					{ __asm volatile ("wfi"); }

////////////// global var begin ///////////////////////////////////////////////
uint32_t SystemCoreClock = 96000000;			// System Clock Frequency (Core Clock) - Startvalue
uint32_t ClockAnpassung=0;						// falls ClockRate schon mal augerufen, dann muss nur noch ein Teil ausgefhrt werden.
int debLevel=2;									// DebugLevel: 0=nothing; 1=error; 2=warn/values; 3=debuginfo
////////////// global var end /////////////////////////////////////////////////

////////////// prototype begin ////////////////////////////////////////////////
void 		ClockRateIRC(uint32_t cpuclk);		// internal RC-Oscillator with 3/6/12/24/48/96 MHz, PLL1 disabled
uint32_t	ClockRateEOsc(uint32_t newspeed);	// external Oscillator, mostly works (but unfortunately not every time ^^) for 12, 24, 48 and 96 MHz
void		PowerOff(uint32_t a);				// a=1=Sleep, 2=DeepSleep, 3=PowerDown, 4=DeepPowerDown
////////////// prototype end //////////////////////////////////////////////////


void ClockRateIRC(uint32_t cpuclk){			// nur definiert fr 3/6/12/24/48/96 MHz, PLL1 deaktiviert
// Interner RC-Osci bei 4MHz, wir wollen 12/24/48/96 MHz out => idealer PLL0-Takt ist 288MHz; damit ist N=1, M=(288*N)/(4*2)=36
// mit PLL0CFG=((N-1)<<16)|(M-1)=0x23
// bei PLL0=276 ergeben sich: 
// Taktteilung knnen wir dann ohne PLL0-nderung ber CCLKCFG erledigen
// PLL1 ist beim Start theoretisch deaktiviert, in der system_LPC17xx.c wird'se aber hochgefahren -> wir haun das Ding weg
 *CCLKCFG= 96;								// sicherheitshalber erst den Clock-Divider runter
 if(ClockAnpassung==0){
  *CLKSRCSEL=0;								// interner RC-Osci
  __disable_irq();
  *PLL1CON=1;								// PLL1 disconnect
  *PLL1FEED=0xAA;*PLL1FEED=0x55;
  while((*PLL1STAT&(1<<25))!=0);
  *PLL1CON=0;								// PLL1 turn off
  *PLL1FEED=0xAA;*PLL1FEED=0x55;
  while((*PLL1STAT&(1<<24))!=0);
  *PLL0CON=1;								// PLL0 disconnect
  *PLL0FEED=0xAA;*PLL0FEED=0x55;
  while((*PLL0STAT&(1<<25))!=0);
  *PLL0CON=0;								// PLL0 turn off
  *PLL0FEED=0xAA;*PLL0FEED=0x55;
  while((*PLL0STAT&(1<<24))!=0);
  *PLL0CFG=0x23;							// ab Bit 0 Multiplizierer (6..512), ab Bit 16 Teiler (1..32), siehe page 37
  *PLL0FEED=0xAA;*PLL0FEED=0x55;
 }
 switch(cpuclk){							// CPU-Takt=PLL0-Takt/CCLKCFG
  case  2: *CCLKCFG=144;*FLASHCFG=0x0000003A;break;
  case  3: *CCLKCFG= 96;*FLASHCFG=0x0000003A;break;
  case  4: *CCLKCFG= 72;*FLASHCFG=0x0000003A;break;
  case  6: *CCLKCFG= 48;*FLASHCFG=0x0000003A;break;
  case 12: *CCLKCFG= 24;*FLASHCFG=0x0000003A;break;
  case 24: *CCLKCFG= 12;*FLASHCFG=0x0000103A;break;
  case 48: *CCLKCFG=  6;*FLASHCFG=0x0001003A;break;
  case 96: *CCLKCFG=  3;*FLASHCFG=0x0010003A;break;
  default: *CCLKCFG= 96;*FLASHCFG=0x0000003A;			// 3MHz ist default
 }
 if(ClockAnpassung==0){
  *PLL0CON=1;								// PLL0 turn on bzw enable
  *PLL0FEED=0xAA;*PLL0FEED=0x55;
  while(!(*PLL0STAT&(1<<26)));
  *PLL0CON=3;								// PLL0 connect
  *PLL0FEED=0xAA;*PLL0FEED=0x55;
  while (!(*PLL0STAT&((1<<25)|(1<<24))));
  __enable_irq();
  ClockAnpassung=1;
 }
}

uint32_t ClockRateEOsc(uint32_t newspeed){	// klappt meist (aber leider nicht immer ^^) fr 12, 24, 48 und 96 MHz
 uint sysclock;              // target clock speed
 uint waits;                 // flash wait time
 uint extclock;              // external clock speed
 uint pllmult;               // PLL multiplier (M)
 uint plldivi;               // PLL divider (N)
 uint plloutd;               // PLL output divider (D)
 uint x=0;					 // Zhler zum Verlassen der while-Schleifen

 extclock=12000000;		//chipCrystal in mbed.ld
 if ((uint)newspeed>=96000000) sysclock=96000000;
 else if ((uint)newspeed>=72000000) sysclock=72000000;
 else if ((uint)newspeed>=48000000) sysclock=48000000;
 else sysclock=(uint)newspeed;
//    LPC_SC->CLKSRCSEL=CLKPWR_CLKSRCSEL_CLKSRC_MAINOSC; // source=main oscillator
 plldivi=4;                               // N
 pllmult=288000000 * 4 / (2 * extclock);  // M
 plloutd=288000000 / sysclock;            // D (output divider)
 
 __disable_irq();

 *CCLKCFG=plloutd-1;                         // output divider
 *PLL0CFG=((plldivi-1)<<16) | (pllmult-1);   // configure

 sysclock=2*pllmult*extclock/(plldivi*plloutd);     // what will result

 // start up PLL0 as per documented sequence
 *PLL0FEED=0xAA;
 *PLL0FEED=0x55;
 *PLL0CON=0x01;
 *PLL0FEED=0xAA;
 *PLL0FEED=0x55;
 // wait for PLL lock
 while (!((*PLL0STAT&(1<<26)))&&(x<5000)){
  x++;
  __enable_irq();
  return 1;
 }

 // connect PLL to output and wait for stable
 *PLL0CON=0x01 | 0x02;
 *PLL0FEED  = 0xAA;
 *PLL0FEED  = 0x55;
 // Wait for PLLC0_STAT & PLLE0_STAT (unequated)
 // [NB: NXP code appears to be wrong for this]
 #define STATBITS (1<<24 | 1<<25)
 x=0;
 while (((*PLL0STAT & STATBITS) != STATBITS)&&(x<5000)){
  x++;
  __enable_irq();
  return 2;
 }

 if (sysclock<=20000000) waits=0;
 else if (sysclock<=40000000) waits=1;
 else if (sysclock<=60000000) waits=2;
 else if (sysclock<=80000000) waits=3;
 else if (sysclock<=100000000) waits=4;
 else waits=5;                        // 'safe'
 // 3A here always required
 *FLASHCFG=0x3A | (waits<<12);  // configure
 SystemCoreClock=sysclock;
  __enable_irq();
 return 0;
}

void PowerOff(uint32_t a){					// a=1=Sleep, 2=DeepSleep, 3=PowerDown, 4=DeepPowerDown
 uint32_t p;								// PCON Power Modes in Bit 0/1; in Bits 8-11 sind Flags fr den SleepMode

 switch(a){
  case 1: p=0;*SCR=0;break;					// SLEEPDEEP-Bit ist *SCR=1<<2;
  case 2: p=0;*SCR=1<<2;break;				// evtl p=8 ? (siehe lpc17xx_clkpwr.c , Fkt CLKPWR_DeepSleep)
  case 3: p=1;*SCR=1<<2;break;				// evtl p=9 ?
  case 4: p=3;*SCR=1<<2;break;
  default: p=0;*SCR=0;;
 }

 *PINMODE0=0xC0FFFFFF;						// alle Pins von PullUp auf PullDown umschalten
 *PINMODE1=0x003FFFFF;
 *PINMODE2=0xF03F030F;
 *PINMODE3=0xFFFFFFFF;
 *PINMODE4=0x0FFFFFFF;
 *PINMODE7=0x003C0000;
 *PINMODE9=0x0F000000;

 *PCON=p;
 while(1){									// bei Bedarf hier eine Aufweckvariable rein
  __asm("nop");								// Pipeline leeren, da Sleep sonst nicht immer klappt
  __asm("nop");
  __asm("nop");
  __asm("wfi");
 }
}

#endif
